/* glwin.h - 25 Jul 09
   Hosts3D - 3D Real-Time Network Monitor
   Copyright (c) 2006-2009  Del Castle

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details. */

#include <GL/freeglut.h>  //GLenum

#include "llist.h"

//selection value
#define GLWIN_OK      0x01
#define GLWIN_ADD     0x02
#define GLWIN_DEL     0x03
#define GLWIN_CLOSE   0x04
#define GLWIN_RESIZE  0x05
#define GLWIN_TEXT    0x06
#define GLWIN_TITLE   0x07
#define GLWIN_UP      0x08
#define GLWIN_DOWN    0x09
#define GLWIN_LIST1   0x0A
#define GLWIN_LIST2   0x0B
#define GLWIN_LIST3   0x0C
#define GLWIN_LIST4   0x0D
#define GLWIN_LIST5   0x0E
#define GLWIN_MNUITM  0x0F
#define GLWIN_MISC1   0x10
#define GLWIN_MISC2   0x11

//2D GUI
class MyGLWin
{
  private:
    struct glwn_obj  //window object
    {
      unsigned char type;  //GLWIN_WINDOW
      bool close, resize;  //display close, resize object
      int name, psel, left, top, width, height, minW, minH;  //minimum width, height
      char title[64];  //title bar text
    } *lastWin;  //last window created
    struct glbt_obj  //button object
    {
      unsigned char type;  //GLWIN_BUTTON
      glwn_obj *pwn;  //parent window object
      bool defalt;  //default
      int left, top, width, value;  //selection value
      char text[16];
    };
    struct glin_obj  //input object
    {
      unsigned char type;  //GLWIN_INPUT
      glwn_obj *pwn;  //parent window object
      bool lower;  //lowercase
      int name, left, top, width;
      unsigned int cursor, first, max;  //cursor position, first char to show, maximum length of text
      char text[256];  //default text
    } *lastInput;  //last input created
    struct gllb_obj  //label object
    {
      unsigned char type;  //GLWIN_LABEL
      glwn_obj *pwn;  //parent window object
      int name, left, top;
      char text[64];
    };
    struct glls_obj  //list object
    {
      unsigned char type;  //GLWIN_LIST
      glwn_obj *pwn;  //parent window object
      glin_obj *lin;  //linked input object
      bool more;  //more list items to follow
      int name, left, top, width, line;  //first line shown
      char items[5][256], fname[256];  //file to list
    };
    struct glvw_obj  //view object
    {
      unsigned char type;  //GLWIN_VIEW
      glwn_obj *pwn;  //parent window object
      bool more;  //more text to follow
      int name, left, top, right, bottom, tab, line;  //first line shown
      char fname[256];  //file to view
    };
    struct glmu_obj  //menu object
    {
      unsigned char type;  //GLWIN_MENU
      int left, top, width, sub, value;  //display sub-menu object
      char text[64];
    };
    int names, selected, currInput, currScroll, mouseX, mouseY, menuY;  //input object with focus (for keys), scroll object with focus (for mouse wheel)
    MyLL GLWinLL;
    void Reset();
    bool MouseOver(int left, int top, int right, int bottom);
    void DrawWin(GLenum mode, glwn_obj *wn);
    void DrawButton(GLenum mode, glbt_obj *bt);
    bool DrawInput(GLenum mode, glin_obj *in);
    void DrawLabel(gllb_obj *lb);
    void DrawList(GLenum mode, glls_obj *ls);
    void DrawView(GLenum mode, glvw_obj *vw);
    void DrawMenu(GLenum mode, glmu_obj *mu);
    void TabInput(bool nx);
    void GLWinDestroy(glwn_obj *pw);
  public:
    MyGLWin();
    ~MyGLWin();
    bool On();
    void CreateWin(int left, int top, int width, int height, const char *title, bool close, bool resize);
    void AddButton(int left, int top, int width, const char *text, bool defalt, int value);
    int AddInput(int left, int top, int width, const char *text, bool lower, int max);
    int AddLabel(int left, int top, const char *text);
    void AddList(int left, int top, int width, const char *fl);
    int AddView(int left, int top, int right, int bottom, int tab, const char *fl);
    void AddMenu(int left, int top, int width, const char *text, int itms, int sub, int value);
    bool Draw(GLenum mode);
    char *GetInput(int name = 0);
    void PutInput(const char *text, int name = 0);
    void PutLabel(const char *text, int name);
    void ResetView(int name);
    int GetSelected();
    void MousePos(int x, int y);
    void Motion(int x, int y);
    void Scroll(int gs);
    void Close(bool all = true);
    int Select(int state, int x, int y);
    void Key(unsigned char key);
};
