/* objects.h - 02 Sep 09
   Hosts3D - 3D Real-Time Network Monitor
   Copyright (c) 2006-2009  Del Castle

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details. */

#include "common.h"

#define HSD_EDTNAME  0x01  //identify edit hostname
#define HSD_EDTRMKS  0x02  //identify edit host remarks
#define HSD_EXPTCSV  0x03  //identify export CSV
#define HSD_FNDHSTS  0x04  //identify find hosts
#define HSD_HNLOPEN  0x05  //identify open network layout file
#define HSD_HNLSAVE  0x06  //identify save network layout file
#define HSD_HPTOPEN  0x07  //identify open packet traffic file
#define HSD_HPTSAVE  0x08  //identify save packet traffic file
#define HSD_HSTINFO  0x09  //identify host info
#define HSD_NETPOS   0x0A  //identify edit netpos
#define HSD_PKTSPRO  0x0B  //identify packets protocol
#define HSD_PKTSPRT  0x0C  //identify packets port
#define HSD_SETCMDS  0x0D  //identify set commands
#define HSD_SLINACT  0x0E  //identify select inactive
#ifndef __MINGW32__
#define HSD_HSENRUN  0x0F  //identify run hsen
#define HSD_HSENSTP  0x10  //identify stop hsen
#endif

const int SPC = 32, MOV = 3, HPR = 20, SELBUF = 50000, SVCS = 32;  //host spacing, movement spacing, hosts per row, host selection buffer, services per host
const float DEPTH = 8192.0;  //perseptive depth
#ifdef __MINGW32__
const char HSDDATA[] = "hsd-data/"  //hosts3d data directory (win)
#else
const char HSDDATA[] = ".hosts3d/"  //hosts3d data directory (linux)
#endif
  , tipn[3][6] = {"IP", "IP/Nm", "Nm"}, tips[3][4] = {"Off", "Sel", "All"}, tona[5][6] = {"Off", "Alert", "IP/Nm", "Host", "Sel"}
  , tclr[10][8] = {"default", "orange", "yellow", "fluro", "green", "mint", "aqua", "blue", "purple", "violet"};

//GL position
struct pos_type
{
  double x, y, z;
};

//GL vertex
struct vtx_type
{
  float x, y, z;
};

//GL triangle
struct tri_type
{
  unsigned char a, b, c;
};

//GL square
struct sqr_type
{
  unsigned char a, b, c, d;
};

//GL view direction
struct view_type
{
  double ax, ay;  //angle
  pos_type ee, dr;  //eye, direction
};

//GL cross object
struct cobj_type
{
  vtx_type vtx[28];
  sqr_type sqr[8];
};

//GL host object
struct hobj_type
{
  vtx_type vtx[10];
  tri_type tri[16];
};

//GL multiple-hosts object
struct mobj_type
{
  vtx_type vtx[11];
  tri_type tri[16];
};

//GL packet object
struct pobj_type
{
  vtx_type vtx[8];
  tri_type tri[12];
};

//host object info
struct host_type
{
  unsigned char lsn, clr, vis, sld:1, pip:1, anm:1, shp:1, lck:1, spr:3;  //last sensor, colour, visible, selected, persistant IP, anomaly, show packets, lock, spare (future use)
  int px, py, pz;
  unsigned long long dld, uld;  //downloads, uploads
  time_t lpk;  //last packet
  host_type *col;
  in_addr hip;
  char htip[16], htmc[18], htnm[256], htrm[256];  //IP, MAC, name, remarks
  int svc[SVCS];  //service
};

//line object info
struct line_type
{
  host_type *sht, *dht;  //source, destination host
  unsigned char spr;  //spare (future use)
};

enum aobj_type { sphr, cube, text };  //alert object - sphere, cube, text

//alert object info
struct alrt_type
{
  aobj_type ao;  //sphere, cube or text alert object
  unsigned char pr, sz;  //size/duration
  host_type *ht;
};

//packet object info
struct pckt_type
{
  unsigned char pr;
  unsigned short dstpt;
  pos_type cu;  //current
  host_type *ht;  //destination host
};

enum ptrc_type { stp, hlt, rec, rpy };  //packet traffic - stop, halt, record, replay

//packet traffic record
struct pkrc_type
{
  timeval ptime;  //time of packet
  pkif_type pk;  //packet
};

enum sipn_type { ips, ins, nms };  //display - IP, IP/name, name only
enum sips_type { off, sel, all };  //IP/name display - off, selection, all
enum sona_type { don, alt, ipn, hst, slt };  //on-active - do nothing, alert, IP/name, host, select
enum sshp_type { non, som, evy };  //show packets - none, some, every

//hosts3d settings
struct sett_type
{
  unsigned char sen, pr, mvd, osd:1, bct:1, fsp:1, ads:1, anm:1, lla:1, pdp:1, spr:1;  //display packets from hsen (0 for all), display packets by protocol (0 for all), master visibility duration, OSD, broadcasts, fast packets, add dests, anomaly detection, automatic link lines, display packet port, spare (future use)
  unsigned short prt;  //display packets by port (0 for all)
  unsigned int pks;  //number of packets allowed on screen before drop
  char clip[256], hssd[128], hsid[4], hsif[256], hshd[16], hssp[256], cmd[4][256];  //input box text clipboard
  sipn_type sipn;  //display IP/name
  sips_type sips;  //IP/name display
  sona_type sona;  //on-active
  sshp_type sshp;  //show packets
  view_type vws[5];
};

//net position/colour
struct netp_type
{
  int px, py, pz;
  char clr, nt[19];  //colour, CIDR net
};

//draw cross object
void cobjDraw();

//draw host object
void hobjDraw(int r, int g, int b);

//draw multiple-hosts object
void mobjDraw(bool sel);

//draw packet object
void pobjDraw(int c);

//return path with hosts3d data directory
char *hsddata(const char *fl);

//add service to host
void svcAdd(host_type *sht, int svc, bool anm);

//destroy net positions LL
void netpsDestroy();

//load netpos file into memory, create empty
void netpsLoad();

//check if a host is to be positioned into a net
bool hostNet(host_type *ht);

//copy packet traffic file from/to
void pkrcCopy(const char *ifn, const char *ofn);

//check for controls file, create it
void checkControls();
